import { autoinject, }                                        from "aurelia-dependency-injection";
import { EventAggregator, Subscription }                      from "aurelia-event-aggregator";
import { Container, LogManager, viewResources }               from "aurelia-framework";
import { PLATFORM }                                           from "aurelia-pal";
import { Router, RouterEvent }                                from "aurelia-router";
import { bindable }                                           from "aurelia-typed-observable-plugin";
import { createElement }                                      from "../utility/utils";
import { NavigationPipelineProccessingResult, RouteSettings } from "./pitaya-app.d";

const META_THEME_COLOR_SELECTOR: string       = "meta[name=theme-color]";
const META_THEME_COLOR_ELEMENT_STRING: string = "<meta name='theme-color'>";

const logger = LogManager.getLogger( "Pitaya:PitayaApp" );
let currentInstance: PitayaApp;

const enum ViewAnimations
{
	lowerToHigher = "pitaya-app--transition-lower-to-higher",
	higherToLower = "pitaya-app--transition-higher-to-lower",
	equalToEqual  = "pitaya-app--transition-equal-to-equal"
}

@autoinject
@viewResources( PLATFORM.moduleName( "./pitaya-app.css" ) )
export class PitayaApp
{
	public static get instance(): PitayaApp
	{
		return currentInstance;
	}

	@bindable public onOnline: Function;
	@bindable public onOffline: Function;
	@bindable public onAttached: Function;
	@bindable public noInteraction: boolean;
	@bindable public statusBarColor: string = "#FFFFFF";

	private _transitionAnimationClass: ViewAnimations = ViewAnimations.equalToEqual;
	private _isOnline: boolean                        = navigator.onLine;
	private _themeColorMetaElement                    = document.querySelector( META_THEME_COLOR_SELECTOR ) || ( document.head.appendChild( createElement( META_THEME_COLOR_ELEMENT_STRING ) ) as HTMLMetaElement );
	private _scopedOnlineMethod                       = this._onOnline.bind( this );
	private _scopedOfflineMethod                      = this._onOffline.bind( this );
	private _routeProcessingSubscribtion: Subscription;

	get container(): Container
	{
		return this._container;
	}

	constructor(
		private _container: Container,
		private _eventAggregator: EventAggregator,
		private _router: Router
	)
	{}

	public async navigate( routeName: string, options?: any ): Promise<boolean>
	{
		const result = this._router.navigate( routeName, options );

		await new Promise( resolve => this._eventAggregator.subscribeOnce( RouterEvent.Complete, resolve ) );

		return result;
	}

	public async navigateToRoute( routeName: string, params?: object, options?: any ): Promise<boolean>
	{
		const result = this._router.navigateToRoute( routeName, params, options );

		await new Promise( resolve => this._eventAggregator.subscribeOnce( RouterEvent.Complete, resolve ) );

		return result;
	}

	public async navigateBack(): Promise<void>
	{
		this._router.navigateBack();

		await new Promise( resolve => this._eventAggregator.subscribeOnce( RouterEvent.Complete, resolve ) );
	}

	public async reinitialize(): Promise<void>
	{
		await this.unbind();
		await this.created();
		await this.bind();
		await this.attached();
	}

	private statusBarColorChanged( value: string )
	{
		this._themeColorMetaElement.setAttribute( "value", value );
	}

	private created()
	{
		if(currentInstance)
		{
			throw new Error("ERROR  [Pitaya:PitayaApp]  It is not allowed to have multiple {PitayaApp}/<pitaya-app> custom elements at a time. You have first to detach the current one from the DOM.",);
		}

		currentInstance = this;

		this._routeProcessingSubscribtion = this._eventAggregator.subscribe( RouterEvent.Processing, this._handleRouteProcessing.bind( this ) );
	}

	private bind()
	{
		window.addEventListener( "online", this._scopedOnlineMethod );
		window.addEventListener( "offline", this._scopedOfflineMethod );
	}

	private attached()
	{
		this.onAttached && this.onAttached( { component : this } );
	}

	private unbind()
	{
		this._routeProcessingSubscribtion.dispose();

		window.removeEventListener( "online", this._scopedOnlineMethod );
		window.removeEventListener( "offline", this._scopedOfflineMethod );

		currentInstance = undefined;
	}

	private async _onOnline()
	{
		this._isOnline = true;

		this.onOnline && this.onOnline( { component : this } );
	}

	private async _onOffline()
	{
		this._isOnline = false;

		this.onOffline && this.onOffline( { component : this } );
	}

	private _handleRouteProcessing( { instruction }: NavigationPipelineProccessingResult )
	{
		if( !instruction.previousInstruction )
		{
			return;
		}

		const previousLevel = ( ( instruction.previousInstruction.config.navModel.settings as RouteSettings ).level || 0 );
		const nextLevel     = ( ( instruction.config.navModel.settings as RouteSettings ).level || 0 );

		if( previousLevel > nextLevel )
		{
			return this._transitionAnimationClass = ViewAnimations.lowerToHigher;
		}
		else if( previousLevel < nextLevel )
		{
			return this._transitionAnimationClass = ViewAnimations.higherToLower;
		}

		this._transitionAnimationClass = ViewAnimations.equalToEqual;
	}
}

export function navigate( routeName: string, options?: any ): Promise<boolean>
{
	return PitayaApp.instance.navigate( routeName, options );
}

export function navigateToRoute( routeName: string, params?: object, options?: any ): Promise<boolean>
{
	return PitayaApp.instance.navigateToRoute( routeName, params, options );
}

export function navigateBack(): Promise<void>
{
	return PitayaApp.instance.navigateBack();
}