import { HttpClient }                                        from "aurelia-fetch-client";
import { Container, LogManager, PLATFORM }                   from "aurelia-framework";
import { RedirectToRoute, RouteConfig, RouterConfiguration } from "aurelia-router";
import { PitayaConfiguration }                               from "../configuration";
import {
	AppDecorator,
	OnlyIfDecorator,
	OnlyIfFunction,
	OnlyIfFunctions,
	OnlyIfStringParameter,
	RedirectConfig,
	RouterDecoratorConfig,
	ViewDecorator
}                                                            from "./decorators.d";

const httpClient = new HttpClient();
const logger     = LogManager.getLogger( "Pitaya:Decorators" );

export const onlyIfFunctions: OnlyIfFunctions = {
	"online"()
	{
		return navigator.onLine;
	},
	"offline"()
	{
		return !navigator.onLine;
	},
	"under development"()
	{
		return Container.instance.get( PitayaConfiguration ).development;
	}
};

export const onlyIf: OnlyIfDecorator = {
	instanceHasProperties( ...instanceProperties: string[] )
	{
		return function( target: object, propertyKey: string, descriptor: PropertyDescriptor ): any
		{
			const oldValue = descriptor.value;

			descriptor.value = function()
			{
				const hasProperty = instanceProperties.some( arg => this.hasOwnProperty( arg ) );

				if( !hasProperty )
				{
					return;
				}

				return oldValue.apply( this, arguments );
			};

			return descriptor;
		};
	},
	is( condition: OnlyIfFunction | OnlyIfStringParameter ): any
	{
		return function( target: object, propertyKey: string, descriptor: PropertyDescriptor ): any
		{
			const oldValue = descriptor.value;

			descriptor.value = function( ...args: any ): any
			{
				try
				{
					const callable = typeof condition === "function" ? condition : onlyIfFunctions[ condition ];

					/* Condition not fulfilled */
					if( !callable.call( this, this ) )
					{
						return;
					}

					return oldValue.call( this, ...args );
				}
				catch( error )
				{
					logger.error( `While trying to apply onlyIf.is() decorator on method '${propertyKey}()'.` );
				}
			};

			return descriptor;
		};
	}
};

export const view: ViewDecorator = {
	redirect( routeConfig: RedirectConfig, condition: ( instance?: unknown ) => boolean )
	{
		return function( constructor: any ): any
		{
			return class extends constructor
			{
				private canActivate()
				{
					if( !condition.call( this, this ) )
					{
						return true;
					}

					const { route, params, options } = routeConfig;

					return new RedirectToRoute( route, params, options );
				}
			};
		};
	}
};

export const app: AppDecorator = {
	router( { title, routesConfig, unknownRoutes, callback }: RouterDecoratorConfig )
	{
		return function( constructor: any ): any
		{
			return class extends constructor
			{
				private async configureRouter( config: RouterConfiguration ): Promise<void>
				{
					const routes    = await getRoutes( routesConfig );
					const mainRoute = routes.find( route => route.name === unknownRoutes ) || routes[ 0 ];

					config.title = title;
					// config.addAuthorizeStep( AuthorizeStep );
					config.map( routes );
					config.mapUnknownRoutes( mainRoute.moduleId );

					callback && callback( config );
				}
			};
		};
	}
};

const getRoutes = async function( routesConfig: string | RouteConfig[] ): Promise<RouteConfig[]>
{
	if( typeof routesConfig === "string" )
	{
		const result     = await httpClient.fetch( PLATFORM.moduleName( routesConfig ) );
		const { routes } = await result.json();

		return routes;
	}

	return routesConfig;
};